"""Placeholders for various elements in output"""

from typing import Iterable, Sized
from colorama import Fore, Style

HL_OPEN = "``"
HL_CLOSE = "''"

HIGHLIGHTS = ("color", "quotes")


def highlight(message: str, highlights: str) -> str:
    """Replace highlight placeholders in a given string using selected method"""
    if highlights == "color":
        return message.replace(HL_OPEN, Fore.RED).replace(HL_CLOSE, Style.RESET_ALL)
    elif highlights == "quotes":
        return message.replace(HL_OPEN, "'").replace(HL_CLOSE, "'")

    raise ValueError("Highlight should be one of: " + str(HIGHLIGHTS))


def hl(what: str) -> str:  # pylint: disable=invalid-name
    """Return highlighted string"""
    return f"{HL_OPEN}{what}{HL_CLOSE}"


def pluralize(name: str, items: Sized) -> str:
    """Return a pluralized name unless there is exactly one element in container."""
    return f"{name}" if len(items) == 1 else f"{name}s"


def hlistify(container: Iterable) -> str:
    """Returns a comma separated list of highlighted names."""
    return ", ".join([hl(name) for name in container])



"""This module contains code that handles comparing function implementations"""

import ast
import logging
from types import MappingProxyType
from typing import Union, List, Optional, Set, FrozenSet, Dict, Mapping
from pyff.kitchensink import hl, pluralize, hlistify
import pyff.imports as pi
import pyff.functions as pf


LOGGER = logging.getLogger(__name__)


class LocalBaseClass:
    # pylint: disable=too-few-public-methods
    """Represents part of a class summary for case when base class is local"""

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"local {hl(self.name)}"


class ImportedBaseClass:
    # pylint: disable=too-few-public-methods
    """Represents part of a class summary for case when base class is imported"""

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"imported {hl(self.name)}"


BaseClassType = Union[LocalBaseClass, ImportedBaseClass]  # pylint: disable=invalid-name


class ClassSummary:  # pylint: disable=too-few-public-methods
    """Contains summary information about a class"""

    def __init__(
        self,
        methods: Set[str],
        definition: ast.ClassDef,
        attributes: Set[str],
        baseclasses: Optional[List[BaseClassType]] = None,
    ) -> None:
        self.methods: Set[str] = methods
        self.attributes = frozenset(attributes)
        self.baseclasses: Optional[List[BaseClassType]] = baseclasses
        self.definition = definition

    @property
    def name(self) -> str:
        """Returns class name"""
        return self.definition.name

    @property
    def public_methods(self) -> FrozenSet[str]:
        """Return public methods of the class"""
        return frozenset({method for method in self.methods if not method.startswith("_")})

    @property
    def private_methods(self) -> FrozenSet[str]:
        """Return public methods of the class"""
        return frozenset({method for method in self.methods if method.startswith("_")})

    def __str__(self) -> str:
        LOGGER.debug("String: %s", repr(self))
        class_part: str = f"class {hl(self.name)}"
        methods = pluralize("method", self.public_methods)
        method_part: str = f"with {len(self.public_methods)} public {methods}"

        if not self.baseclasses:
            return f"{class_part} {method_part}"
        elif len(self.baseclasses) == 1:
            return f"{class_part} derived from {str(self.baseclasses[0])} {method_part}"

        raise Exception("Multiple inheritance not yet implemented")

    def __lt__(self, other):
        return self.name < other.name

    def __repr__(self):
        return (
            f"ClassSummary(methods={self.methods}, attributes={self.attributes}, "
            f"bases={self.baseclasses}, AST={self.definition}"
        )


class ClassesExtractor(ast.NodeVisitor):
    """Extracts information about classes in a module"""

    class SelfAttributeExtractor(ast.NodeVisitor):
        """Extracts self attributes references used in the node"""

        def __init__(self):
            self.attributes: Set[str] = set()

        def visit_Attribute(self, node):  # pylint: disable=invalid-name
            """self.attribute -> 'attribute'"""
            if isinstance(node.value, ast.Name) and node.value.id == "self":
                self.attributes.add(node.attr)

    class AssignmentExtractor(ast.NodeVisitor):
        """Extracts self attributes used as assignment targets"""

        def __init__(self):
            self.attributes: Set[str] = set()
            self.extractor = ClassesExtractor.SelfAttributeExtractor()

        def visit_Assign(self, node):  # pylint: disable=invalid-name
            """'self.attribute = value' -> 'attribute'"""
            for target in node.targets:
                self.extractor.visit(target)
            self.attributes.update(self.extractor.attributes)

        def visit_AnnAssign(self, node):  # pylint: disable=invalid-name
            """'self.attribute: typehint = value' -> 'attribute'"""
            self.extractor.visit(node.target)
            self.attributes.update(self.extractor.attributes)

    class MethodExtractor(ast.NodeVisitor):
        """Extracts information about a method"""

        @staticmethod
        def extract_attributes(node: ast.FunctionDef) -> FrozenSet[str]:
            """Extract attributes used in the method"""
            LOGGER.debug("Extracting attributes from method '%s", node.name)
            extractor = ClassesExtractor.AssignmentExtractor()
            for statement in node.body:
                extractor.visit(statement)

            LOGGER.debug("Discovered attributes '%s'", extractor.attributes)
            return frozenset(extractor.attributes)

        def __init__(self):
            self.methods: Set[str] = set()
            self.attributes: Set[str] = set()

        def visit_FunctionDef(self, node):  # pylint: disable=invalid-name
            """Save counts of encountered private/public methods"""
            LOGGER.debug("Extracting information about method '%s'", node.name)
            self.methods.add(node.name)
            self.attributes.update(self.extract_attributes(node))

    def __init__(self, names: Optional[pi.ImportedNames] = None) -> None:
        self._classes: Dict[str, ClassSummary] = {}
        self._names: Optional[pi.ImportedNames] = names

    @property
    def classes(self) -> Mapping[str, ClassSummary]:
        """Return a set of extracted class summaries"""
        return MappingProxyType(self._classes)

    @property
    def classnames(self) -> FrozenSet[str]:
        """Return a set of class names in the module"""
        return frozenset(self._classes.keys())

    def visit_ClassDef(self, node):  # pylint: disable=invalid-name
        """Save information about classes that appeared in a module"""
        LOGGER.debug("Extracting information about class '%s'", node.name)

        extractor = ClassesExtractor.MethodExtractor()
        extractor.visit(node)

        bases: List[str] = []
        for base in node.bases:
            if base.id in self._names:
                LOGGER.debug("Imported ancestor class '%s', base.id")
                bases.append(ImportedBaseClass(base.id))
            else:
                LOGGER.debug("Local ancestor class '%s', base.id")
                bases.append(LocalBaseClass(base.id))

        summary = ClassSummary(
            extractor.methods, baseclasses=bases, definition=node, attributes=extractor.attributes
        )
        self._classes[node.name] = summary


class AttributesPyfference:  # pylint: disable=too-few-public-methods
    """Represents differnces between attributes of two classes"""

    def __init__(self, removed: FrozenSet[str], new: FrozenSet[str]) -> None:
        self.removed: FrozenSet[str] = removed
        self.new: FrozenSet[str] = new

    def __str__(self):
        lines: List[str] = []
        if self.removed:
            lines.append(
                f"Removed {pluralize('attribute', self.removed)} {hlistify(sorted(self.removed))}"
            )
        if self.new:
            lines.append(f"New {pluralize('attribute', self.new)} {hlistify(sorted(self.new))}")

        return "\n".join(lines)

    def __bool__(self) -> bool:
        return bool(self.removed or self.new)


class ClassPyfference:  # pylint: disable=too-few-public-methods
    """Represents differences between two classes"""

    def __init__(
        self,
        name: str,
        attributes: Optional[AttributesPyfference],
        methods: Optional[pf.FunctionsPyfference],
    ) -> None:
        self.attributes: Optional[AttributesPyfference] = attributes
        self.methods: Optional[pf.FunctionsPyfference] = methods
        self.name = name

    def __str__(self):
        lines = [f"Class {hl(self.name)} changed:"]
        if self.attributes:
            lines.append(str(self.attributes))
        if self.methods:
            self.methods.set_method()
            methods = str(self.methods)
            lines.append(methods)

        return "\n".join(lines).replace("\n", "\n  ")


class ClassesPyfference:  # pylint: disable=too-few-public-methods
    """Holds differences between classes defined in a module"""

    def __init__(self, new: Set[ClassSummary], changed: Dict[str, ClassPyfference]) -> None:
        self.new: Set[ClassSummary] = new
        self.changed: Dict[str, ClassPyfference] = changed

    def __str__(self):
        new = [f"New {cls}" for cls in sorted(self.new)]
        changed = [str(self.changed[name]) for name in sorted(self.changed)]
        return "\n".join(changed + new)

    def simplify(self) -> Optional["ClassesPyfference"]:
        """Cleans empty differences, empty sets etc. after manipulation"""
        return self if self.new else None


def pyff_class(
    old: ClassSummary,
    new: ClassSummary,
    old_imports: pi.ImportedNames,
    new_imports: pi.ImportedNames,
) -> Optional[ClassPyfference]:
    """Return differences in two classes"""
    methods = pf.pyff_functions(old.definition, new.definition, old_imports, new_imports)
    attributes = AttributesPyfference(
        removed=old.attributes - new.attributes, new=new.attributes - old.attributes
    )
    if methods or attributes:
        LOGGER.debug(f"Class '{new.name}' differs")
        return ClassPyfference(name=new.name, methods=methods, attributes=attributes)

    LOGGER.debug(f"Class '{old.name}' is identical")
    return None


def pyff_classes(
    old: ast.Module, new: ast.Module, old_imports: pi.ImportedNames, new_imports: pi.ImportedNames
) -> Optional[ClassesPyfference]:
    """Return differences in classes defined in two modules"""
    first_walker = ClassesExtractor(names=old_imports)
    second_walker = ClassesExtractor(names=new_imports)

    first_walker.visit(old)
    second_walker.visit(new)

    differences: Dict[str, ClassPyfference] = {}
    both = first_walker.classnames.intersection(second_walker.classnames)
    LOGGER.debug(f"Classes present in both module versions: {both}")
    for klass in both:
        LOGGER.debug(f"Comparing class '{klass}'")
        difference = pyff_class(
            first_walker.classes[klass], second_walker.classes[klass], old_imports, new_imports
        )
        LOGGER.debug(f"Difference: {difference}")
        if difference:
            LOGGER.debug(f"Class {klass} differs")
            differences[klass] = difference
        else:
            LOGGER.debug(f"Class {klass} is identical")

    new_classes = {
        cls for cls in second_walker.classes.values() if cls.name not in first_walker.classnames
    }
    LOGGER.debug(f"New classes: {new_classes}")

    if differences or new_classes:
        LOGGER.debug("Classes differ")
        return ClassesPyfference(new=new_classes, changed=differences)

    LOGGER.debug("Classes are identical")
    return None



"""This module contains code that handles comparing revisions in Git repository"""

import tempfile
import shutil
import pathlib
from typing import Optional
import git


import pyff.directories as pd
import pyff.packages as pp


class RevisionsPyfference:  # pylint: disable=too-few-public-methods
    """Represents a difference between two revisions"""

    # RevisionsPyfference is basically the same as DirectoryPyfference, so we
    # embed DirectoryPyfference and delegate everything to it
    def __init__(self, change: pd.DirectoryPyfference) -> None:
        self._change: pd.DirectoryPyfference = change

    def __str__(self):
        return str(self._change)

    @property
    def packages(self) -> Optional[pp.PackagesPyfference]:
        """Return what Python packages differ between revisions"""
        return self._change.packages


def pyff_git_revision(repository: str, old: str, new: str) -> Optional[RevisionsPyfference]:
    """Compare two revisions in a Git repository"""
    with tempfile.TemporaryDirectory() as temporary_wd:
        working_directory = pathlib.Path(temporary_wd)
        source_dir = working_directory / "source"
        old_dir = working_directory / "old"
        new_dir = working_directory / "new"

        repo = git.Repo.clone_from(repository, source_dir)

        repo.git.checkout(old)
        shutil.copytree(source_dir, old_dir)

        repo.git.checkout(new)
        shutil.copytree(source_dir, new_dir)

        change = pd.pyff_directory(old_dir, new_dir)
        if change:
            return RevisionsPyfference(change)

        return None



"""This module contains code that handles comparing packages"""
import logging
import pathlib
import ast

from typing import Optional, Iterable, FrozenSet, Set, Dict, Mapping
from types import MappingProxyType
from astroid.modutils import get_module_files

import pyff.modules as pm
from pyff.kitchensink import hl, hlistify, pluralize

LOGGER = logging.getLogger(__name__)


class PackageSummary:  # pylint: disable=too-few-public-methods
    """Holds information about a Python package"""

    def __init__(self, package: pathlib.Path) -> None:
        self.path = package


class PackagePyfference:  # pylint: disable=too-few-public-methods
    """Holds differences between two Python packages"""

    def __init__(self, modules: Optional[pm.ModulesPyfference]) -> None:
        self.modules: Optional[pm.ModulesPyfference] = modules

    def __str__(self):
        return str(self.modules)

    def __repr__(self):
        return f"PackagePyfference(modules={repr(self.modules)})"


class PackagesPyfference:  # pylint: disable=too-few-public-methods
    """Holds differences between packages in package or directory"""

    def __init__(
        self,
        removed: Dict[pathlib.Path, PackageSummary],
        changed: Dict[pathlib.Path, PackagePyfference],
        new: Dict[pathlib.Path, PackageSummary],
    ) -> None:
        self._removed: Dict[pathlib.Path, PackageSummary] = removed
        self._changed: Dict[pathlib.Path, PackagePyfference] = changed
        self._new: Dict[pathlib.Path, PackageSummary] = new

    @property
    def removed(self) -> Mapping[pathlib.Path, PackageSummary]:
        """Read-only view on removed packages"""
        return MappingProxyType(self._removed)

    @property
    def new(self) -> Mapping[pathlib.Path, PackageSummary]:
        """Read-only view on new packages"""
        return MappingProxyType(self._new)

    @property
    def changed(self) -> Mapping[pathlib.Path, PackagePyfference]:
        """Read-only view on changed packages"""
        return MappingProxyType(self._changed)

    def __str__(self):
        lines = []
        if self._removed:
            lines.append(f"Removed {pluralize('package', self._removed)} {hlistify(self._removed)}")

        if self._changed:
            lines.append(
                "\n".join(
                    [
                        f"Package {hl(package)} changed:\n  " + str(change).replace("\n", "\n  ")
                        for package, change in self._changed.items()
                    ]
                )
            )

        if self._new:
            lines.append(f"New {pluralize('package', self._new)} {hlistify(self._new)}")

        return "\n".join(lines)

    def __bool__(self):
        return bool(self._removed or self._changed or self._new)


def extract_modules(files: Iterable[pathlib.Path], package: PackageSummary) -> FrozenSet[str]:
    """Extract direct modules of a packages (i.e. not modules of subpackages"""
    return frozenset({module.name for module in files if module.parents[0] == package.path})


def _compare_module_in_packages(
    module: pathlib.Path, old_package: PackageSummary, new_package: PackageSummary
) -> Optional[pm.ModulePyfference]:
    """Compare one module in two packages"""
    old_module = old_package.path / module
    new_module = new_package.path / module

    return pm.pyff_module_code(old_module.read_text(), new_module.read_text())


def summarize_package(package: pathlib.Path) -> PackageSummary:
    """Create a PackageSummary for a given path"""
    return PackageSummary(package)


def _summarize_module_in_package(module: pathlib.Path, package: PackageSummary) -> pm.ModuleSummary:
    full_path = package.path / module
    module_ast = ast.parse(full_path.read_text())
    return pm.ModuleSummary(str(module), module_ast)


def pyff_package(
    old_package: PackageSummary, new_package: PackageSummary
) -> Optional[PackagePyfference]:
    """Given summaries of two versions of a package, return differences between them"""
    old_files: Set[pathlib.Path] = {
        pathlib.Path(module) for module in get_module_files(old_package.path, ())
    }
    new_files: Set[pathlib.Path] = {
        pathlib.Path(module) for module in get_module_files(new_package.path, ())
    }

    LOGGER.debug("Files of the old package %s: %s", str(old_package.path), old_files)
    LOGGER.debug("Files of the new package %s: %s", str(new_package.path), new_files)

    old_modules: Set[pathlib.Path] = {
        pathlib.Path(module) for module in extract_modules(old_files, old_package)
    }
    new_modules: Set[pathlib.Path] = {
        pathlib.Path(module) for module in extract_modules(new_files, new_package)
    }

    LOGGER.debug("Old modules: %s", str(old_modules))
    LOGGER.debug("New modules: %s", str(new_modules))

    removed = old_modules - new_modules
    new = new_modules - old_modules
    both = old_modules.intersection(new_modules)

    LOGGER.debug("Removed modules: %s", str(removed))
    LOGGER.debug("Modules in both packages: %s", str(both))
    LOGGER.debug("New modules: %s", str(new))

    removed_summaries = {
        str(module): _summarize_module_in_package(module, old_package) for module in removed
    }
    new_summaries = {
        str(module): _summarize_module_in_package(module, new_package) for module in new
    }
    changed = {
        str(module): change
        for module, change in [
            (module, _compare_module_in_packages(module, old_package, new_package))
            for module in both
        ]
        if change is not None
    }
    modules = pm.ModulesPyfference(removed_summaries, changed, new_summaries)

    if modules:
        return PackagePyfference(modules)

    return None


def pyff_package_path(old: pathlib.Path, new: pathlib.Path) -> Optional[PackagePyfference]:
    """Given *paths* to two versions of a package, return differences between them"""
    old_summary = summarize_package(old)
    new_summary = summarize_package(new)
    return pyff_package(old_summary, new_summary)



"""This module containst code that handles comparing function implementations"""

import logging
import pathlib
from typing import Optional, Tuple, FrozenSet, Set

import pyff.packages as pp


LOGGER = logging.getLogger(__name__)


class DirectoryPyfference:  # pylint: disable=too-few-public-methods
    """Represents differences between two directories"""

    def __init__(self, packages: Optional[pp.PackagesPyfference]) -> None:
        self.packages: Optional[pp.PackagesPyfference] = packages

    def __str__(self):
        output = []
        if self.packages:
            output.append(str(self.packages))

        return "\n".join(output)

    def __bool__(self):
        return bool(self.packages)


def find_those_pythonz(
    directory: pathlib.Path
) -> Tuple[FrozenSet[pathlib.Path], FrozenSet[pathlib.Path]]:
    """Find Python packages and modules in a given directory"""
    traverse_directories: Set[pathlib.Path] = {directory}
    packages: Set[pathlib.Path] = set()
    modules: Set[pathlib.Path] = set()
    while traverse_directories:
        candidate = traverse_directories.pop()
        LOGGER.debug("Checking directory for Python content: %s", candidate)
        possible_init = candidate / "__init__.py"
        if possible_init.exists():
            LOGGER.debug("Directory %s is a Python package", candidate)
            packages.add(candidate.relative_to(directory))
            # We do not want to dig into packages
            continue

        for item in candidate.iterdir():
            if item.is_file() and item.suffix == ".py":
                LOGGER.debug("File %s is a Python module", item)
                modules.add(item.relative_to(directory))
            elif item.is_dir():
                LOGGER.debug("Scheduling directory '%s' for traversal", item)
                traverse_directories.add(item)
            else:
                LOGGER.debug("Item %s is not anything Python-related, so we do not care", str(item))

    return (frozenset(packages), frozenset(modules))


def pyff_directory(old: pathlib.Path, new: pathlib.Path) -> Optional[DirectoryPyfference]:
    """Find Python packages and modules in two directories and compare them"""
    if not (old.is_dir() and new.is_dir()):
        raise ValueError(f"At least one of {old}, {new} is not an existing directory")

    old_pkgs, old_modules = find_those_pythonz(old)
    new_pkgs, new_modules = find_those_pythonz(new)

    LOGGER.debug("Packages in the old directory: %s", str(old_pkgs))
    LOGGER.debug("Packages in the new directory: %s", str(new_pkgs))
    LOGGER.debug("Modules in the old directory: %s", str(old_modules))
    LOGGER.debug("Modules in the new directory: %s", str(new_modules))

    removed_packages = old_pkgs - new_pkgs
    both_packages = old_pkgs.intersection(new_pkgs)
    new_packages = new_pkgs - old_pkgs

    LOGGER.debug("Removed packages: %s", str(removed_packages))
    LOGGER.debug("Packages in both directories: %s", str(both_packages))
    LOGGER.debug("New packages: %s", str(new_packages))

    removed_package_summaries = {pkg: pp.summarize_package(old / pkg) for pkg in removed_packages}
    changed_packages = {
        pkg: change
        for pkg, change in [
            (pkg, pp.pyff_package(pp.summarize_package(old / pkg), pp.summarize_package(new / pkg)))
            for pkg in both_packages
        ]
        if change is not None
    }
    new_package_summaries = {pkg: pp.summarize_package(new / pkg) for pkg in new_packages}

    if removed_package_summaries or changed_packages or new_package_summaries:
        return DirectoryPyfference(
            packages=pp.PackagesPyfference(
                removed_package_summaries, changed_packages, new_package_summaries
            )
        )

    return None



"""This module contains code that handles comparing function implementations"""

import ast
import logging
from itertools import zip_longest
from typing import Optional, Set, Dict, List, Union, FrozenSet
from collections.abc import Hashable

import pyff.imports as pi
import pyff.statements as ps
from pyff.kitchensink import hl, hlistify


LOGGER = logging.getLogger(__name__)


class FunctionImplementationChange(Hashable):  # pylint: disable=too-few-public-methods
    """Represents any single change in function implementation"""

    def make_message(self) -> str:  # pylint: disable=no-self-use
        """Returns a human-readable message explaining the change"""
        return f"Code semantics changed"

    def __eq__(self, other):
        return isinstance(other, FunctionImplementationChange)

    def __hash__(self):
        return hash("FunctionImplementationChange")

    def __repr__(self):
        return "FunctionImplementationChange"


class ExternalUsageChange(FunctionImplementationChange):  # pylint: disable=too-few-public-methods
    """Represents any change in how function uses external names"""

    def __init__(self, gone: Set[str], appeared: Set[str]) -> None:
        self.gone: Set[str] = gone
        self.appeared: Set[str] = appeared

    def make_message(self) -> str:
        """Returns a human-readable message explaining the change"""
        lines: List[str] = []
        if self.gone:
            lines.append(f"No longer uses imported {hlistify(sorted(self.gone))}")
        if self.appeared:
            lines.append(f"Newly uses imported {hlistify(sorted(self.appeared))}")

        return "\n".join(lines)

    def __eq__(self, other):
        return self.gone == other.gone and self.appeared == other.appeared

    def __hash__(self):
        return hash("ExternalUsageChange")

    def __repr__(self):
        return f"ExternalUsageChange(gone={self.gone}, appeared={self.appeared})"


class StatementChange(FunctionImplementationChange):  # pylint: disable=too-few-public-methods
    """Represents a change between two statements"""

    def __init__(self, change: ps.StatementPyfference) -> None:
        self.change: ps.StatementPyfference = change

    def make_message(self) -> str:
        return str(self.change)

    def __hash__(self):
        return hash("StatementChange")

    def __repr__(self):
        return f"StatementChange(change={self.change})"


class FunctionPyfference:  # pylint: disable=too-few-public-methods
    """Holds differences between Python function definitions"""

    def __init__(
        self,
        name: str,
        implementation: Set[FunctionImplementationChange],
        old_name: Optional[str] = None,
    ) -> None:
        self.name: str = name
        self.old_name: Optional[str] = old_name
        self.implementation: Set[FunctionImplementationChange] = implementation
        self._noun: str = "Function"

    def __str__(self) -> str:
        lines: List[str] = []
        if self.old_name is not None:
            lines.append(f"{self._noun} {hl(self.old_name)} renamed to {hl(self.name)}")

        if self.implementation:
            lines.append(f"{self._noun} {hl(self.name)} changed implementation:")

        implementation_changes = []
        for change in self.implementation:
            implementation_changes.append("  " + change.make_message().replace("\n", "\n  "))
        lines.extend(sorted(implementation_changes))

        return "\n".join(lines)

    def set_method(self):
        """Used when FunctionPyfference is used in context of a class"""
        self._noun = "Method"

    def simplify(self) -> Optional["FunctionPyfference"]:
        """Cleans empty differences, empty sets etc. after manipulation"""
        return self if self.old_name or self.implementation else None

    def __repr__(self):
        return (
            f"FunctionPyfference(name={self.name}, implementation={self.implementation}, "
            f"old_name={self.old_name}, noun={self._noun})"
        )


class FunctionSummary:  # pylint: disable=too-few-public-methods
    """Contains summary information about a function"""

    def __init__(self, name: str, node: ast.FunctionDef, is_property: bool = False) -> None:
        self.node: ast.FunctionDef = node
        self.name: str = name
        self.property: bool = is_property
        self._noun: str = "function"

    def __eq__(self, other):
        return self.name == other.name

    def set_method(self):
        """Used when FunctionPyfference is used in context of a class"""
        self._noun = "method"

    def __str__(self):
        prop = "property " if self.property else ""
        return f"{prop}{self._noun} {hl(self.name)}"

    def __repr__(self):
        return (
            f"FunctionSummary(name={self.name}, node={repr(self.node)}, property={self.property})"
        )


class FunctionPyfferenceRecorder:
    """Records various changes detected in function definitions."""

    def __init__(self, name: str) -> None:
        self.name: str = name
        self.old_name: Optional[str] = None
        self.implementation: Set[FunctionImplementationChange] = set()

    def name_changed(self, old_name: str) -> None:
        """Record a name change of the function."""
        self.old_name = old_name

    def implementation_changed(self, change: FunctionImplementationChange) -> None:
        """Record changes in implementation of the function."""
        self.implementation.add(change)

    def build(self) -> Optional[FunctionPyfference]:
        """Produces final FunctionPyfference from recorded changes.

        If no changes were recorded, return None."""
        if self.old_name is None and not self.implementation:
            return None

        return FunctionPyfference(
            name=self.name, old_name=self.old_name, implementation=self.implementation
        )


class ExternalNamesExtractor(ast.NodeVisitor):
    """Collects information about imported name usage in function"""

    def __init__(self, imported_names: pi.ImportedNames) -> None:
        self.imported_names: pi.ImportedNames = imported_names
        self.names: Set[str] = set()
        self.in_progress: Optional[str] = None

    def visit_Name(self, node):  # pylint: disable=invalid-name
        """Compare all names against a list of imported names"""
        self.in_progress = None
        self.generic_visit(node)
        if node.id in self.imported_names:
            self.names.add(node.id)
        else:
            self.in_progress = node.id

    def visit_Attribute(self, node):  # pylint: disable=invalid-name
        """..."""
        self.in_progress = None
        self.generic_visit(node)
        if self.in_progress is not None:
            self.in_progress = ".".join((self.in_progress, node.attr))
            if self.in_progress in self.imported_names:
                self.names.add(self.in_progress)
                self.in_progress = None


def compare_import_usage(  # pylint: disable=invalid-name
    old: ast.FunctionDef,
    new: ast.FunctionDef,
    old_imports: pi.ImportedNames,
    new_imports: pi.ImportedNames,
) -> Optional[ExternalUsageChange]:
    """Compares external (imported) names usage in two versions of a function.

    Args:
        old: Old version of the function
        new: New version of the function
        old_imports: Imported names available for old version of the function
        new_imports: Imported names available for new version of the function

    Returns:
        ExternalUsageChange object, which is basically a pair of two sets: `appeared`, which
        contains names only used in new version of the function, and `gone`, which contains
        names only used in the old version. If both sets would be empty, None is returned."""

    first_walker = ExternalNamesExtractor(old_imports)
    second_walker = ExternalNamesExtractor(new_imports)

    for statement in old.body:
        first_walker.visit(statement)

    for statement in new.body:
        second_walker.visit(statement)

    appeared = second_walker.names - first_walker.names
    gone = first_walker.names - second_walker.names

    LOGGER.debug(f"Imported names used in old function: {first_walker.names}")
    LOGGER.debug(f"Imported names used in new function: {second_walker.names}")
    LOGGER.debug(f"Imported names not used anymore:     {gone}")
    LOGGER.debug(f"Imported names newly used:           {appeared}")

    if appeared or gone:
        return ExternalUsageChange(gone=gone, appeared=appeared)

    return None


def pyff_function(
    old: FunctionSummary,
    new: FunctionSummary,
    old_imports: pi.ImportedNames,
    new_imports: pi.ImportedNames,
) -> Optional[FunctionPyfference]:
    """Return differences between two Python functions.

    Args:
        old: Old version of Python function.
        new: New version of Python function.
        old_imports: Imported names available for old version of the function
        new_imports: Imported names available for new version of the function

    Returns:
        If the functions are identical, returns None. If they differ, a FunctionPyfference
        object is returned, describing the differences."""

    difference_recorder = FunctionPyfferenceRecorder(new.name)

    if old.name != new.name:
        LOGGER.debug(f"Name differs: old={old.name} new={new.name}")
        difference_recorder.name_changed(old.name)

    for old_statement, new_statement in zip_longest(old.node.body, new.node.body):
        if old_statement is None or new_statement is None:
            LOGGER.debug(f"  old={repr(old_statement)}")
            LOGGER.debug(f"  new={repr(new_statement)}")
            LOGGER.debug(
                "  One statement is None: one function is longer, so implementation changed"
            )
            difference_recorder.implementation_changed(FunctionImplementationChange())
            break

        change = ps.pyff_statement(old_statement, new_statement, old_imports, new_imports)
        if change:
            LOGGER.debug("  Statements are different")
            LOGGER.debug(f"  old={ast.dump(old_statement)}")
            LOGGER.debug(f"  new={ast.dump(new_statement)}")
            LOGGER.debug(f"  change={repr(change)}")
            if change.is_specific():
                difference_recorder.implementation_changed(StatementChange(change))
            else:
                difference_recorder.implementation_changed(FunctionImplementationChange())

    LOGGER.debug("Comparing imported name usage")
    external_name_usage_difference = compare_import_usage(
        old.node, new.node, old_imports, new_imports
    )
    if external_name_usage_difference:
        LOGGER.debug("Imported name usage differs")
        difference_recorder.implementation_changed(external_name_usage_difference)

    return difference_recorder.build()


def pyff_function_code(
    old: str, new: str, old_imports: pi.ImportedNames, new_imports: pi.ImportedNames
) -> Optional[FunctionPyfference]:
    """Return differences between two Python functions.

    Args:
        old: Old version of Python function.
        new: New version of Python function.
        old_imports: Imported names available for old version of the function
        new_imports: Imported names available for new version of the function

    Returns:
        If the functions are identical, returns None. If they differ, a FunctionPyfference
        object is returned, describing the differences."""

    extractor = FunctionsExtractor()
    try:
        extractor.visit(ast.parse(old))
        old_summary = extractor.functions.popitem()[1]
    except KeyError:
        raise ValueError("Old module does not seem to contain exactly one function code")

    try:
        extractor.visit(ast.parse(new))
        new_summary = extractor.functions.popitem()[1]
    except KeyError:
        raise ValueError("Old module does not seem to contain exactly one function code")

    return pyff_function(old_summary, new_summary, old_imports, new_imports)


class FunctionsExtractor(ast.NodeVisitor):
    """Extract information about functions in a module"""

    def __init__(self) -> None:
        self.functions: Dict[str, FunctionSummary] = {}

    def visit_ClassDef(self, node):  # pylint: disable=invalid-name
        """Prevent this visitor from inspecting classes"""
        pass

    @property
    def names(self) -> FrozenSet[str]:
        """Returns a set of names of discovered functions"""
        return frozenset(self.functions.keys())

    @staticmethod
    def _is_property_decorator(node: Union[ast.Name, ast.Attribute]) -> bool:
        return isinstance(node, ast.Name) and node.id == "property"

    def visit_FunctionDef(self, node):  # pylint: disable=invalid-name
        """Save top-level function definitions"""
        is_property: bool = False

        for decorator in node.decorator_list:
            if self._is_property_decorator(decorator):
                is_property = True
                break

        self.functions[node.name] = FunctionSummary(
            name=node.name, node=node, is_property=is_property
        )


class FunctionsPyfference:  # pylint: disable=too-few-public-methods
    """Holds differences between top-level functions in a module"""

    def __init__(
        self,
        new: Dict[str, FunctionSummary],
        changed: Dict[str, FunctionPyfference],
        removed: Dict[str, FunctionSummary],
    ) -> None:
        self.changed: Dict[str, FunctionPyfference] = changed
        self.new: Dict[str, FunctionSummary] = new
        self.removed: Dict[str, FunctionSummary] = removed

    def __str__(self) -> str:
        removed = "\n".join(
            [f"Removed {f}" for f in sorted([str(self.removed[name]) for name in self.removed])]
        )
        changed = "\n".join([str(self.changed[name]) for name in sorted(self.changed)])
        new = "\n".join([f"New {f}" for f in sorted([str(self.new[name]) for name in self.new])])

        return "\n".join([changeset for changeset in (removed, changed, new) if changeset])

    def set_method(self):
        """Used when FunctionPyfference is used in context of a class"""
        for function in self.removed.values():
            function.set_method()
        for function in self.changed.values():
            function.set_method()
        for function in self.new.values():
            function.set_method()

    def simplify(self) -> Optional["FunctionsPyfference"]:
        """Cleans empty differences, empty sets etc. after manipulation"""
        new_changed = {}
        for function, change in self.changed.items():
            new_change = change.simplify()
            if new_change:
                new_changed[function] = new_change
        self.changed = new_changed

        return self if self.new or self.changed or self.removed else None


def pyff_functions(
    old: Union[ast.Module, ast.ClassDef],
    new: Union[ast.Module, ast.ClassDef],
    old_imports: pi.ImportedNames,
    new_imports: pi.ImportedNames,
) -> Optional[FunctionsPyfference]:
    """Return differences in top-level functions in two modules"""
    old_walker = FunctionsExtractor()
    new_walker = FunctionsExtractor()

    for node in old.body:
        old_walker.visit(node)

    for node in new.body:
        new_walker.visit(node)

    both: FrozenSet[str] = old_walker.names.intersection(new_walker.names)
    LOGGER.debug(f"Functions present in both modules: {both}")
    differences: Dict[str, FunctionPyfference] = {}
    for function in both:
        LOGGER.debug(f"Comparing function '{function}'")
        difference = pyff_function(
            old_walker.functions[function], new_walker.functions[function], old_imports, new_imports
        )
        LOGGER.debug(f"Difference: {repr(difference)}")
        if difference:
            LOGGER.debug(f"Function {function} differs")
            differences[function] = difference
        else:
            LOGGER.debug(f"Function {function} is identical")

    new_names: FrozenSet[str] = new_walker.names - old_walker.names
    new_functions: Dict[str, FunctionSummary] = {
        name: new_walker.functions[name] for name in new_names
    }
    LOGGER.debug(f"New functions: {new_names}")

    removed_names: FrozenSet[str] = old_walker.names - new_walker.names
    removed_functions: Dict[str, FunctionSummary] = {
        name: old_walker.functions[name] for name in removed_names
    }
    LOGGER.debug(f"Removed functions: {removed_names}")

    if differences or new_functions or removed_functions:
        LOGGER.debug("Functions differ")
        return FunctionsPyfference(
            removed=removed_functions, changed=differences, new=new_functions
        )

    LOGGER.debug("Functions are identical")
    return None



"""Python diff"""

import setuptools
import pyff

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setuptools.setup(
    name="pythondiff",
    version=pyff.__version__,
    description="Python Diff",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/petr-muller/pyff",
    author="Petr Muller",
    author_email="afri@afri.cz",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Quality Assurance",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
    ],
    keywords="python static_analysis diff",
    packages=["pyff"],
    setup_requires=["pytest-runner", "pytest-bdd", "pytest-pylint", "pytest-mypy", "pytest-cov"],
    tests_require=["pytest", "pylint", "mypy"],
    install_requires=["colorama", "astroid", "gitpython"],
    entry_points={
        "console_scripts": [
            "pyff=pyff.run:pyffmod",
            "pyff-package=pyff.run:pyffpkg",
            "pyff-dir=pyff.run:pyffdir",
            "pyff-git=pyff.run:pyffgit",
        ]
    },
)
